{ File   : DS9Ext.dpr
  Author : Deniz Oezmen
  Created: 2005-08-17, based on DS9EXTR.PAS created ~2001
  Changed: 2006-08-02

  Star Trek: DS9 Harbinger CAT/ALL/LST extractor.
}
program DS9Ext;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TBuffer = array of Byte;

var
  AllFile: file;
  CatFile: Text;
  LstFile: file of Longword;
  CatOpen: Boolean;

function IntToStrL(i: Longint; Len: Byte): string;
begin
  Result := IntToStr(i);
  while Length(Result) < Len do
    Result := '0' + Result
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenFiles(FileName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(AllFile, FileName + '.all');

  {$I-}
  Reset(AllFile, 1);
  {$I+}
  Result := (IOResult = 0);
  if not Result then
  begin
    WriteLn('Error: Could not open', FileName + '.all');
    Exit;
  end;

  AssignFile(CatFile, FileName + '.cat');

  {$I-}
  Reset(CatFile);
  {$I+}
  CatOpen := (IOResult = 0);

  AssignFile(LstFile, FileName + '.lst');

  {$I-}
  Reset(LstFile);
  {$I+}
  Result := (IOResult = 0);
  if not Result then
    WriteLn('Error: Could not open', FileName + '.lat')
end;

function CloseFiles: Longint;
begin
  {$I-}
  Close(AllFile);
  Close(CatFile);
  Close(LstFile);
  {$I+}
  Result := IOResult
end;

procedure ExtractFiles(FileName: string);
var
  i, j,
  Size,
  NumEntries: Longword;
  DestFile  : file;
  Line      : string;
  Offsets   : array of Longword;
  Buffer    : TBuffer;
begin
  // get number of entries according to .lst file
  NumEntries := Longword(FileSize(LstFile)) - 1;

  if not CatOpen then
  begin
    WriteLn('Warning: No .cat file found. Auto-generating file names.');
    FileName := Prefix(FileName)
  end
  else
  begin
    // counter-check with .cat entries
    ReadLn(CatFile, Line);
    if Longword(StrToInt(Line)) <> NumEntries then
    begin
      WriteLn(
        'Error: Number of entries inconsistent in .cat and .lst files!');
      WriteLn('Aborting.');
      Exit
    end;
  end;

  WriteLn(FileName, ': ', NumEntries, ' entries');

  // get offset list
  SetLength(Offsets, NumEntries + 1);
  BlockRead(LstFile, Offsets[0], NumEntries + 1);

  // scan through the list
  for i := 0 to NumEntries - 1 do
  begin
    // get file name and open file for write access
    if CatOpen then
      ReadLn(CatFile, Line)
    else
      Line := FileName + '#' + IntToStrL(i, 5);

    FileMode := fmOpenWrite;
    AssignFile(DestFile, Line);
    Rewrite(DestFile, 1);

    // go to read offset and compute file size
    Seek(AllFile, Offsets[i]);
    j := i + 1;
    repeat
      Size := Offsets[j] - Offsets[i];
      Inc(j)
    until (Size <> 0) or (j = NumEntries);

    WriteLn(' => ', Line, ' @ ', Offsets[i], ' (', Size, ')');

    // copy contents
    SetLength(Buffer, Size);
    BlockRead(AllFile, Buffer[0], Size);
    BlockWrite(DestFile, Buffer[0], Size);
    Close(DestFile)
  end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  ALLName   : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  FilePara := FilePara + '.all';
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara 
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      ALLName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenFiles(Copy(ALLName, 1, Length(AllName) - 4)) then
      begin
        ExtractFiles(Copy(ALLName, 1, Length(AllName) - 4));
        CloseFiles
      end
      else
        Writeln(ALLName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found');
  FreeMem(PFilePara, Length(FilePara) + 1)
end;

begin
  if ParamCount < 1 then
  begin
    WriteLn('Usage: DS9Extr <file>');
    WriteLn('  Opens <file>.all, <file>.cat, <file>.lst');
  end
  else
    ParseFiles(ParamStr(1))
end.
